// Copyright 2017 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
// Modified by Jeremy Retailleau.

#define TF_MAX_ARITY 7
#include <pxr/arch/defines.h>
#if defined(ARCH_OS_LINUX)
#include <unistd.h>
#endif
#include <algorithm>
#include <any>
#include <atomic>
#include <cinttypes>
#include <cmath>
#include <cstdarg>
#include <cstdint>
#include <cstring>
#include <iosfwd>
#include <iostream>
#include <list>
#include <locale>
#include <map>
#include <math.h>
#include <memory>
#include <optional>
#include <set>
#include <sstream>
#include <stdarg.h>
#include <stddef.h>
#include <string>
#include <sys/types.h>
#include <tuple>
#include <type_traits>
#include <typeindex>
#include <typeinfo>
#include <utility>
#include <variant>
#include <vector>
